from _typeshed import Incomplete as Incomplete

logger: Incomplete

def Plot_estimability(round_data, path, solver) -> None: ...
def plot_rCC_vs_k(x_values, rCC_values, round, framework_settings, solver) -> None: ...
def plot_sobol_results(time_samples, sobol_analysis_results, sobol_problem, solver, response_key, framework_settings) -> None: ...

class Plotting_Results:
    path: Incomplete
    case: Incomplete
    mutation_settings: Incomplete
    modelling_folder: str
    def __init__(self, modelling_settings, framework_settings) -> None: ...
    def fit_plot(self, data, result, round, model_structure) -> None: ...
    def conf_plot(self, parameters, cov_matrices, confidence_intervals, round) -> None: ...

class Plotting_FinalResults:
    round_data: Incomplete
    winner_solver: Incomplete
    selected_rounds: Incomplete
    def __init__(self, round_data, winner_solver, selected_rounds) -> None: ...
    def conf_plot(self, filename, ellipsoid_volume_filename: str = 'ellipsoid_volume_across_rounds.png', std_devs_filename: str = 'parameter_std_devs_across_rounds.png', parameter_estimates_filename: str = 'parameter_estimates_across_rounds.png'): ...
    def conf_plot_metrics(self, ellipsoid_volume_filename, std_devs_filename) -> None: ...
    def accvsprec_plot(self, filename) -> None: ...
    def pt_plot(self, filename, roundc) -> None: ...
    def reporter(self, filename) -> None: ...
    def pcomp_plot(self, save_path) -> None: ...
    def tcomp_plot(self, save_path) -> None: ...

def validation_R2(prediction_metric, validation_metric, reference_metric, full_path, case) -> None: ...
def validation_params(parameters, ref_params, full_path) -> None: ...
def save_sobol_results_to_excel(results2, folder_path, file_name: str = 'sobol_results.xlsx') -> None: ...



def thetadic05() -> None: ...
def f05(t, y, phi, phit, theta, te) -> None: ...
def thetadic06() -> None: ...
def f06(t, y, phi, phit, theta, te) -> None: ...
def thetadic07() -> None: ...
def f07(t, y, phi, phit, theta, te) -> None: ...
def thetadic11() -> None: ...
def f11(t, y, phi, phit, theta, te) -> None: ...
def thetadic08() -> None: ...
def f08(t, y, phi, phit, theta, te) -> None: ...
def thetadic09() -> None: ...
def f09(t, y, phi, phit, theta, te) -> None: ...
def thetadic12() -> None: ...
def f12(t, y, phi, phit, theta, te) -> None: ...
def thetadic13() -> None: ...
def f13(t, y, phi, phit, theta, te) -> None: ...
def thetadic14() -> None: ...
def f14(t, y, phi, phit, theta, te) -> None: ...
def thetadic14p() -> None: ...
def f14p(t, y, phi, phit, theta, te) -> None: ...
def thetadic14s() -> None: ...
def f14s(t, y, phi, phit, theta, te) -> None: ...
def thetadic15() -> None: ...
def f15(t, y, phi, phit, theta, te) -> None: ...
def thetadic16() -> None: ...
def f16(t, y, phi, phit, theta, te) -> None: ...
def thetadic17() -> None: ...
def f17(t, y, phi, phit, theta, te) -> None: ...
def thetadic22() -> None: ...
def f22(t, y, phi, phit, theta, te) -> None: ...



import numpy as np
from scipy.integrate import solve_ivp
from scipy.interpolate import interp1d
import importlib

def Simula(t, swps, uphi, uphisc, uphitsc, utsc, utheta, uthetac, cvp, uphit, model_name, model_structure, modelling_settings):
    """
    Simulate the system dynamics using the provided parameters and model.

    Parameters
    ----------
    t : list or array of float
        Time points for the simulation.
    swps : dict of str: list
        Dictionary containing time and level vectors for piecewise interpolation.
        Keys are variable names ending with 't' or 'l' that define time and level vectors.
    uphi : dict of str: float
        Time-invariant variables scaled by their maxima.
    uphisc : dict of str: float
        Scaling factors for each time-invariant variable.
    uphitsc : dict of str: float
        Scaling factors for each time-variant variable.
    utsc : float
        Scaling factor for the time points.
    utheta : list of float
        Model parameters (unscaled).
    uthetac : list of float
        Scaling factors for each parameter in utheta.
    cvp : dict of str: str
        A dictionary mapping each time-variant input variable to an interpolation method:
        - 'CPF': Constant piecewise function (previous step interpolation)
        - 'LPF': Linear piecewise function
        - 'none': No interpolation, use a constant from uphit
    uphit : dict of str: float or list
        Initial or constant values for time-variant variables when using the 'none' method.
    model_name : str
        The name of the model function to be used for simulation.
    model_structure : dict
        Defines model structure, including variables, attributes, and initial conditions.
    modelling_settings : dict
        Settings for the modelling process, including external functions.

    Returns
    -------
    tuple
        tv_ophi (dict): Time-varying outputs {var_name: ndarray}
        ti_ophi (dict): Time-invariant outputs (currently empty)
        phit (dict): Scaled piecewise interpolated data {var_name: ndarray}
    """
    # Ensure 'ext_func' key is present in modelling_settings
    if 'ext_func' not in modelling_settings:
        raise KeyError("'ext_func' key is missing in modelling_settings")

    # Check if model_name is in modelling_settings['ext_func']
    if model_name in modelling_settings['ext_func']:
        model = modelling_settings['ext_func'][model_name]
    else:
        # Attempt to import the model from kernel_models if not found in ext_func
        try:
            krnl_models = importlib.import_module('middoe.krnl_models')
            model = getattr(krnl_models, model_name)
        except (ImportError, AttributeError):
            raise KeyError(f"Model function '{model_name}' not found in 'ext_func' or 'kernel_models'")

    tv_ophi, ti_ophi, phit = _ivpsolver(
        t, swps, uphi, uphisc, uphitsc, utsc, utheta, uthetac,
        cvp, uphit, model, model_structure, modelling_settings
    )
    return tv_ophi, ti_ophi, phit

def _Piecewiser(t, swps, cvp, phit):
    """
    Perform piecewise interpolation for trajectories of time-variant input variables.

    Parameters
    ----------
    t : array-like
        Time points at which to evaluate the piecewise functions.
    swps : dict of str: list
        Dictionary containing time and level vectors for piecewise functions.
        For a variable 'X', 'Xt' defines times and 'Xl' defines levels.
    cvp : dict of str: str
        Dictionary of methods for each variable. Keys are variable names, values are:
        - 'CPF': Constant piecewise function (previous step interpolation)
        - 'LPF': Linear piecewise function
        - 'none': No interpolation, use phit values as constants.
    phit : dict of str: float or list
        Initial or constant values for each variable when using the 'none' method.

    Returns
    -------
    dict of str: ndarray
        Dictionary mapping variable names to arrays of interpolated values.
    """

    result = {}

    # Iterate over each variable defined in cvp
    for var, method in cvp.items():
        time_key = f'{var}t'
        level_key = f'{var}l'

        if method == 'CPF':
            # For CPF, we need both time and level data in swps
            if time_key in swps and level_key in swps:
                f_previous = interp1d(swps[time_key], swps[level_key], kind='previous', fill_value="extrapolate")
                result[var] = f_previous(t)
            else:
                raise KeyError(f"Missing keys for CPF method: '{time_key}' or '{level_key}'")

        elif method == 'LPF':
            # For LPF, we need both time and level data in swps
            if time_key in swps and level_key in swps:
                result[var] = np.interp(t, swps[time_key], swps[level_key])
            else:
                raise KeyError(f"Missing keys for LPF method: '{time_key}' or '{level_key}'")

        elif method == 'none':
            # For 'none', we do not rely on swps. Instead, we use the constant value from phit.
            if var not in phit:
                raise KeyError(f"No initial value provided in phit for variable '{var}' under 'none' method.")
            value = phit[var]
            if isinstance(value, (list, np.ndarray)):
                result[var] = np.full_like(t, value[0])
            else:
                result[var] = np.full_like(t, value)

        else:
            # Unrecognized method
            raise ValueError(f"Unrecognized piecewise method '{method}' for variable '{var}'.")

    return result


def _ivpsolver(t, swps, uphi, uphisc, uphitsc, utsc, utheta, uthetac, cvp, uphit, model, model_structure, modelling_settings):
    """
    Solve the ODE system defined by the model using solve_ivp.

    Parameters
    ----------
    t : list or array of float
        Time points for the simulation.
    swps : dict
        Switching points data for piecewise interpolation.
    uphi : dict
        Time-invariant variables (scaled to [0,1]).
    uphisc : dict
        Scaling factors for time-invariant variables.
    uphitsc : dict
        Scaling factors for time-variant variables.
    utsc : float
        Scaling factor for time.
    utheta : list of float
        Model parameters (unscaled).
    uthetac : list of float
        Scaling factors for parameters.
    cvp : dict of str: str
        Dictionary specifying interpolation methods per variable.
    uphit : dict
        Initial or constant values for time-variant variables when using 'none'.
    model : callable
        The model function to be integrated.
    model_structure : dict
        Defines the structure of the model, including variable attributes.
    modelling_settings : dict
        Additional modelling settings.

    Returns
    -------
    tuple
        tv_ophi (dict): {variable_name: ndarray of solution values over time}
        ti_ophi (dict): {time_invariant_output_name: value} (currently empty)
        phit (dict): {variable_name: ndarray of interpolated input values}
    """
    # Scale time-invariant variables by their max values
    phi = {k: v * uphisc[k] for k, v in uphi.items() if k in uphisc}

    # Scale parameters
    theta = [utheta[i] * uthetac[i] for i in range(len(utheta))]

    # Compute piecewise interpolations and scale them
    piecewise_data = _Piecewiser(t, swps, cvp, uphit)
    phit = {k: piecewise_data[k] * uphitsc[k] for k in uphitsc}

    # Construct initial conditions
    y0 = []
    for var in model_structure['tv_ophi'].keys():
        var_attrs = model_structure['tv_ophi'][var]
        initial_val = var_attrs['initials']
        if initial_val == 'variable':
            # Use time-invariant variable with '_0' suffix
            y0.append(phi[var + '_0'])
        else:
            # Use the specified initial value directly
            y0.append(initial_val)

    # Scale the time vector
    t_scaled = (np.array(t) * utsc).tolist()

    # Solve the ODE system
    result = solve_ivp(
        model,
        [min(t_scaled), max(t_scaled)],
        y0,
        method='LSODA',
        t_eval=t_scaled,
        args=(phi, phit, theta, t_scaled)
    )

    # Extract time-varying outputs
    tv_ophi = {f'y{i+1}': result.y[i] for i in range(len(result.y))}
    ti_ophi = {}  # Currently empty, add logic if needed

    return tv_ophi, ti_ophi, phit



from scipy.optimize import minimize
import numpy as np
from scipy.optimize import differential_evolution
from middoe.iden_utils import _initialize_dictionaries

def Parmest(model_structure, modelling_settings, estimation_settings, data, run_solver, case=None):
    """
    Main function to perform parameter estimation.

    Parameters:
    model_structure (dict): User provided - Dictionary containing the model structure.
    modelling_settings (dict): User provided - Dictionary containing the modelling settings.
    estimation_settings (dict): User provided - Dictionary containing the estimation settings.
    data (dict): Dictionary containing the data for optimization, from experiments.
    run_solver (function): Model simulator function.

    Returns:
    tuple: Results of the optimization and theta parameters.
    """

    # Unpack design settings
    _initialize_dictionaries(modelling_settings, estimation_settings)

    active_solvers = modelling_settings['active_solvers']
    theta_parameters = modelling_settings['theta_parameters']
    bound_max = modelling_settings['bound_max']
    bound_min = modelling_settings['bound_min']
    mutation = modelling_settings['mutation']
    method = estimation_settings['method']
    objf = estimation_settings['objf']


    if case is None:
        # Check if 'normalized_parameters' exists in modelling_settings
        if 'normalized_parameters' in modelling_settings:
            x0_dict = modelling_settings['normalized_parameters']
        else:
            x0_dict = estimation_settings['x0']
    elif case == 'freeze':
        x0_dict = estimation_settings['x0']
    logging= estimation_settings['logging']

    # Call runner function
    results = _runner(
        active_solvers, theta_parameters, bound_max, bound_min, mutation,
        objf, x0_dict, method, data, model_structure, run_solver, modelling_settings
    )

    # Report results
    _report_optimization_results(results, logging)

    return results

# def _objective(theta, data, active_solvers, x0, thetac, thetas, model_structure, run_solver, modelling_settings, **kwargs):
#     """
#     Objective function for optimization in parameter estimation.
#
#     Parameters:
#     theta (list): List of parameters to optimize.
#     data (dict): Dictionary containing the data for optimization.
#     active_solvers (list): List of active solvers.
#     x0 (list): Initial guess for the parameters.
#     thetac (list): List of parameter scalors.
#     nd (int): time discrete points.
#     thetas (list): List indicating which parameters are to be optimized.
#     design_variables (dict): Dictionary containing design variables , part of the model structure.
#     run_solver (function): Model simulator function.
#     **kwargs: Additional keyword arguments.
#
#     Returns:
#     tuple: Optimized data and metrics.
#     """
#     if len(theta) != len(x0):
#         raise ValueError("Length of theta and x0 must be the same")
#
#     theta = theta.tolist()
#
#     # Replace elements in theta with x0 where thetas is False
#     theta = [theta[i] if thetas[i] else x0[i] for i in range(len(thetas))]
#
#     global_y_true = {}
#     global_y_pred = {}
#     maxv = {}  # To store max values per sheet and variable
#
#     # Extract variables from design_variables
#     tv_iphi_vars = list(model_structure['tv_iphi'].keys())
#     ti_iphi_vars = list(model_structure['ti_iphi'].keys())
#     tv_ophi_vars = [var for var in model_structure['tv_ophi'].keys() if model_structure['tv_ophi'][var].get('measured', True)]
#     ti_ophi_vars = [var for var in model_structure['ti_ophi'].keys() if model_structure['ti_ophi'][var].get('measured', True)]
#
#     std_dev = {
#         var: model_structure['tv_ophi'][var]['unc']
#         for var in model_structure['tv_ophi'].keys()
#         if model_structure['tv_ophi'][var].get('measured', True)
#     }
#
#     # Initialize scaling factors
#     phisc = {var: 1 for var in ti_iphi_vars}  # Scaling factor for time-invariant inputs
#     phitsc = {var: 1 for var in tv_iphi_vars}  # Scaling factor for time-variant inputs
#     tsc = 1  # Time-scaling factor
#
#     # Loop through data
#     for sheet_name, sheet_data in data.items():
#         t_values = {
#             var: np.array(sheet_data[f"MES_X:{var}"])[~np.isnan(sheet_data[f"MES_X:{var}"])]
#             for var in tv_ophi_vars
#             if f"MES_X:{var}" in sheet_data
#         }
#         # Process time-variant and time-invariant input variables
#         swps_data = {}
#         for var in tv_iphi_vars:
#             if f'{var}t' not in sheet_data or f'{var}l' not in sheet_data:
#                 pass
#             else:
#                 # Filter out NaN values for time ('t') and level ('l') variables
#                 valid_t = np.array(sheet_data[f'{var}t'])[~np.isnan(sheet_data[f'{var}t'])]
#                 valid_l = np.array(sheet_data[f'{var}l'])[~np.isnan(sheet_data[f'{var}l'])]
#                 # Only add to swps_data if both valid_t and valid_l have valid (non-empty) data
#                 if valid_t.size > 0 and valid_l.size > 0:
#                     swps_data[f'{var}t'] = valid_t
#                     swps_data[f'{var}l'] = valid_l
#
#         # Extract input variables (time-invariant and time-variant)
#         ti_iphi_data = {var: np.array(sheet_data.get(var, [np.nan]))[0] for var in ti_iphi_vars}
#         tv_iphi_data = {var: np.array(sheet_data.get(var, []))[~np.isnan(sheet_data.get(var, []))] for var in
#                         tv_iphi_vars}
#
#         # Extract time-variant output variables for measured variables
#         tv_ophi_data = {
#             var: np.array(sheet_data.get(f"MES_Y:{var}", []))[~np.isnan(sheet_data.get(f"MES_Y:{var}", []))]
#             for var in tv_ophi_vars
#             if f"MES_Y:{var}" in sheet_data
#         }
#
#         # Extract time-invariant output variables for measured variables
#         ti_ophi_data = {
#             var: np.array(sheet_data.get(f"MES_Y:{var}", [np.nan]))[0]
#             for var in ti_ophi_vars
#             if f"MES_Y:{var}" in sheet_data
#         }
#
#         # Initialize maxv for this sheet
#         maxv.setdefault(sheet_name, {})
#
#         # Piecewise function check
#         # Construct cvp_doe dynamically for tv_iphi variables
#         cvp = {}
#
#         # Filter tv_iphi variables
#         tv_iphi_vars = model_structure['tv_iphi'].keys()
#
#         # Loop through tv_iphi variables
#         for var in tv_iphi_vars:
#             cvp_col = f"CVP:{var}"  # Expect column like "CVP:u1", "CVP:u2", etc.
#             if cvp_col not in sheet_data:
#                 raise ValueError(f"Missing '{cvp_col}' column in sheet: {sheet_name}")
#             # Assign the value from the sheet data
#             cvp[var] = sheet_data[cvp_col].iloc[0]
#
#         solver_name = active_solvers[0]
#
#         if solver_name:
#             # Ensure the `X:all` column exists in the data
#             if "X:all" not in sheet_data:
#                 raise ValueError(f"Missing 'X:all' column in sheet: {sheet_name}")
#
#             # Read the time data directly from the `X:all` column
#             t_valuese = np.array(sheet_data["X:all"])[~np.isnan(sheet_data["X:all"])]
#
#             # Ensure time values are unique and sorted
#             t_valuese = np.unique(t_valuese)
#
#             # Pass all required arguments to the solver
#             tv_ophi, ti_ophi, phit_interp = run_solver(
#                 t_valuese, swps_data, ti_iphi_data, phisc, phitsc, tsc, theta, thetac, cvp,
#                 tv_iphi_data, solver_name, model_structure, modelling_settings
#             )
#             tv_ophi_filtered = {}
#             indices={}
#             for var in tv_ophi_vars:
#                 if var not in t_values:
#                     raise ValueError(f"No time values found for variable '{var}' in t_values")
#
#                 # Get the specific time values for this variable
#                 t_values_var = t_values[var]
#
#                 # Find indices in the full simulation time (`t_valuese`) that match the specific times for this variable
#                 indices[var] = np.isin(t_valuese, t_values_var)
#
#                 # Filter the corresponding output values in `tv_ophi`
#                 if var in tv_ophi:
#                     tv_ophi_filtered[var] = np.array(tv_ophi[var])[indices[var]]
#                 else:
#                     raise KeyError(f"Variable '{var}' not found in tv_ophi")
#
#             # Collect true and predicted values, and apply scaling per sheet and variable
#             for var in tv_ophi_vars:
#                 # Get true values
#                 y_true = tv_ophi_data.get(var, np.array([]))
#                 # Get predicted values
#                 y_pred = tv_ophi_filtered.get(var, np.zeros_like(y_true))
#
#                 # Ensure matching lengths
#                 min_len = min(len(y_true), len(y_pred))
#                 y_true = y_true[:min_len]
#                 y_pred = y_pred[:min_len]
#
#                 # Find max value between true and predicted for this sheet and variable
#                 max_val = max(np.max(np.abs(y_true)), np.max(np.abs(y_pred)))
#                 if max_val == 0:
#                     max_val = 1  # Avoid division by zero
#
#                 # Store max_val
#                 maxv.setdefault(sheet_name, {})[var] = max_val
#
#                 # Scale the data
#                 y_true_scaled = y_true / max_val
#                 y_pred_scaled = y_pred / max_val
#
#                 # Store scaled data
#                 global_y_true.setdefault(var, []).extend(y_true_scaled)
#                 global_y_pred.setdefault(var, []).extend(y_pred_scaled)
#
#             # For time-invariant variables
#             for var in ti_ophi_vars:
#                 y_true = ti_ophi_data.get(var, np.nan)
#                 y_pred = ti_ophi.get(var, np.nan)
#
#                 if not np.isnan(y_true) and not np.isnan(y_pred):
#                     # Find max value between true and pred
#                     max_val = max(abs(y_true), abs(y_pred))
#                     if max_val == 0:
#                         max_val = 1  # Avoid division by zero
#
#                     # Store max_val
#                     maxv[sheet_name][var] = max_val
#
#                     # Scale the data
#                     y_true_scaled = y_true / max_val
#                     y_pred_scaled = y_pred / max_val
#
#                     # Store scaled data
#                     global_y_true.setdefault(var, []).append(y_true_scaled)
#                     global_y_pred.setdefault(var, []).append(y_pred_scaled)
#
#     # Combine all variables' data into a single array for metrics computation
#     all_y_true = []
#     all_y_pred = []
#
#     # Handle variables
#     for var in global_y_true:
#         y_true = np.array(global_y_true[var])
#         y_pred = np.array(global_y_pred.get(var, np.zeros_like(y_true)))
#
#         # Ensure matching lengths by trimming the longer array
#         min_len = min(len(y_true), len(y_pred))
#         y_true = y_true[:min_len]
#         y_pred = y_pred[:min_len]
#
#         # Append to the combined lists
#         all_y_true.extend(y_true)
#         all_y_pred.extend(y_pred)
#
#     # Convert to NumPy arrays
#     all_y_true = np.array(all_y_true)
#     all_y_pred = np.array(all_y_pred)
#
#     # Ensure both arrays have the same length
#     min_len = min(len(all_y_true), len(all_y_pred))
#     all_y_true = all_y_true[:min_len]
#     all_y_pred = all_y_pred[:min_len]
#
#     # LS
#     ss_res = np.sum((all_y_true - all_y_pred) ** 2)
#     mean_y_true = np.mean(all_y_true)
#     ss_tot = np.sum((all_y_true - mean_y_true) ** 2)
#     LS = 1 - (ss_res / ss_tot)
#     # MLE
#     residuals = all_y_true - all_y_pred
#     sigma_squared_mle = np.mean(residuals ** 2)
#     MLE = (len(all_y_true) / 2) * np.log(2 * np.pi * sigma_squared_mle) + \
#           (1 / (2 * sigma_squared_mle)) * np.sum(residuals ** 2)
#     # Chi
#     denominator = np.abs(all_y_pred)
#     denominator[denominator == 0] = 1e-8  # Avoid division by zero
#     Chi = np.sum((all_y_true - all_y_pred) ** 2 / denominator)
#     # JWLS
#     JWLS = np.dot(residuals.T, residuals)
#
#     # Store metrics for the combined dataset
#     metrics = {
#         'LS': LS,
#         'MLE': MLE,
#         'Chi': Chi,
#         'JWLS': JWLS
#     }
#
#     return data, metrics

def _objective(
    theta,
    data,
    active_solvers,
    x0,
    thetac,
    thetas,
    model_structure,
    run_solver,
    modelling_settings,
    **kwargs
):
    """
    Objective function for optimization in parameter estimation.

    Parameters:
    ----------
    theta : list
        List of parameters to optimize.
    data : dict
        Dictionary containing the data for optimization.
    active_solvers : list
        List of active solvers.
    x0 : list
        Initial guess for the parameters.
    thetac : list
        List of parameter scalars.
    thetas : list
        List indicating which parameters are to be optimized.
    model_structure : dict
        Dictionary that describes the structure of the model
        (variables, measured or not, uncertainties, etc.).
    run_solver : function
        Model simulator function.
    modelling_settings : dict
        Settings for running the model.
    **kwargs : dict
        Additional keyword arguments.

    Returns:
    ----------
    tuple
        (data, metrics)

        data : dict
            The original data (unmodified).
        metrics : dict
            A dictionary containing combined metrics:
            {
                'LS': float,       # classical R² over *all* scaled data
                'MLE': float,      # negative log-likelihood with known sigmas
                'Chi': float,      # Chi-squared
                'JWLS': float,     # Weighted least squares
                'R2_responses': {var1: R²_for_var1, var2: ..., ...}
            }
    """

    if len(theta) != len(x0):
        raise ValueError("Length of theta and x0 must be the same")

    # Convert to list (in case it's a numpy array)
    theta = theta.tolist()

    # Replace elements in theta with x0 where thetas is False
    theta = [theta[i] if thetas[i] else x0[i] for i in range(len(thetas))]

    # -------------------------------------------------------------------------
    # Prepare structures for storing global data (across sheets) that we'll use
    # to compute overall metrics at the end.
    # We'll keep both unscaled (for computing R²) and scaled or weighted sets.
    # -------------------------------------------------------------------------
    # For unweighted global R² across *all* variables (scaled by max-abs):
    global_y_true = {}
    global_y_pred = {}
    maxv = {}  # max absolute value per sheet & variable

    # For computing per-variable R², as well as JWLS and MLE,
    # we want to track each measurement along with its sigma.
    # var_measurements[var] = list of (y_true_i, y_pred_i, sigma_i)
    var_measurements = {}

    # -------------------------------------------------------------------------
    # Extract time-variant and time-invariant measured variables from model_structure
    # -------------------------------------------------------------------------
    # By default, we consider only measured (get('measured', True) == True).
    tv_iphi_vars = list(model_structure['tv_iphi'].keys())
    ti_iphi_vars = list(model_structure['ti_iphi'].keys())
    tv_ophi_vars = [
        var for var in model_structure['tv_ophi'].keys()
        if model_structure['tv_ophi'][var].get('measured', True)
    ]
    ti_ophi_vars = [
        var for var in model_structure['ti_ophi'].keys()
        if model_structure['ti_ophi'][var].get('measured', True)
    ]

    # -------------------------------------------------------------------------
    # Attempt to retrieve standard deviations for each measured variable.
    # If not found, default to 1.0.
    # We'll store them in a dict so that for each variable we know its sigma.
    # Note: The same sigma is used for all time steps for that variable (assuming
    # a single std. dev. per variable).
    # -------------------------------------------------------------------------
    std_dev = {}
    # TV outputs
    for var in model_structure['tv_ophi'].keys():
        if model_structure['tv_ophi'][var].get('measured', True):
            sigma = model_structure['tv_ophi'][var].get('unc', None)
            if sigma is None or np.isnan(sigma):
                sigma = 1.0
            std_dev[var] = sigma
    # TI outputs
    for var in model_structure['ti_ophi'].keys():
        if model_structure['ti_ophi'][var].get('measured', True):
            sigma = model_structure['ti_ophi'][var].get('unc', None)
            if sigma is None or np.isnan(sigma):
                sigma = 1.0
            std_dev[var] = sigma

    # -------------------------------------------------------------------------
    # Initialize scaling factors for time-variant and time-invariant inputs
    # (used when calling the solver).
    # -------------------------------------------------------------------------
    phisc = {var: 1 for var in ti_iphi_vars}   # Scaling for TI inputs
    phitsc = {var: 1 for var in tv_iphi_vars}  # Scaling for TV inputs
    tsc = 1  # Time-scaling factor

    # -------------------------------------------------------------------------
    # Main loop over each sheet of data
    # -------------------------------------------------------------------------
    for sheet_name, sheet_data in data.items():

        # Gather the time vectors for each measured TV output
        # (e.g., "MES_X:var" columns hold the times where that output was measured).
        t_values = {
            var: np.array(sheet_data[f"MES_X:{var}"])[~np.isnan(sheet_data[f"MES_X:{var}"])]
            for var in tv_ophi_vars
            if f"MES_X:{var}" in sheet_data
        }

        # swps_data will hold piecewise time + level data for piecewise functions
        # (time-variant inputs).
        swps_data = {}
        for var in tv_iphi_vars:
            if f'{var}t' not in sheet_data or f'{var}l' not in sheet_data:
                pass
            else:
                valid_t = np.array(sheet_data[f'{var}t'])[~np.isnan(sheet_data[f'{var}t'])]
                valid_l = np.array(sheet_data[f'{var}l'])[~np.isnan(sheet_data[f'{var}l'])]
                if valid_t.size > 0 and valid_l.size > 0:
                    swps_data[f'{var}t'] = valid_t
                    swps_data[f'{var}l'] = valid_l

        # Extract input variables (time-invariant and time-variant)
        ti_iphi_data = {
            var: np.array(sheet_data.get(var, [np.nan]))[0]
            for var in ti_iphi_vars
        }
        tv_iphi_data = {
            var: np.array(sheet_data.get(var, []))[~np.isnan(sheet_data.get(var, []))]
            for var in tv_iphi_vars
        }

        # Extract output variables that are measured (time-variant and time-invariant).
        tv_ophi_data = {
            var: np.array(sheet_data.get(f"MES_Y:{var}", []))[~np.isnan(sheet_data.get(f"MES_Y:{var}", []))]
            for var in tv_ophi_vars
            if f"MES_Y:{var}" in sheet_data
        }
        ti_ophi_data = {
            var: np.array(sheet_data.get(f"MES_Y:{var}", [np.nan]))[0]
            for var in ti_ophi_vars
            if f"MES_Y:{var}" in sheet_data
        }

        # Initialize maxv dict for this sheet
        maxv.setdefault(sheet_name, {})

        # Construct cvp for tv_iphi variables (piecewise control or similar).
        cvp = {}
        for var in model_structure['tv_iphi'].keys():
            cvp_col = f"CVP:{var}"
            if cvp_col not in sheet_data:
                raise ValueError(f"Missing '{cvp_col}' column in sheet: {sheet_name}")
            cvp[var] = sheet_data[cvp_col].iloc[0]

        # We'll use the first solver in active_solvers, or you could loop over them.
        solver_name = active_solvers[0]

        if solver_name:
            if "X:all" not in sheet_data:
                raise ValueError(f"Missing 'X:all' column in sheet: {sheet_name}")

            # Read the time data (for the entire simulation) from "X:all"
            t_valuese = np.array(sheet_data["X:all"])[~np.isnan(sheet_data["X:all"])]
            t_valuese = np.unique(t_valuese)  # ensure unique & sorted

            # Run solver
            tv_ophi_sim, ti_ophi_sim, phit_interp = run_solver(
                t_valuese,
                swps_data,
                ti_iphi_data,
                phisc,
                phitsc,
                tsc,
                theta,
                thetac,
                cvp,
                tv_iphi_data,
                solver_name,
                model_structure,
                modelling_settings
            )

            # Filter solver outputs at the measurement times for each tv_ophi_vars
            tv_ophi_filtered = {}
            indices = {}
            for var in tv_ophi_vars:
                if var not in t_values:
                    raise ValueError(f"No time values found for variable '{var}' in t_values")
                t_values_var = t_values[var]

                # Find indices in the full simulation time that match the
                # measurement times for this variable.
                indices[var] = np.isin(t_valuese, t_values_var)

                if var in tv_ophi_sim:
                    tv_ophi_filtered[var] = np.array(tv_ophi_sim[var])[indices[var]]
                else:
                    raise KeyError(f"Variable '{var}' not found in tv_ophi_sim output")

            # -------------------------------------------------------------
            # Collect measured vs. predicted for TIME-VARIANT outputs
            # -------------------------------------------------------------
            for var in tv_ophi_vars:
                y_true = tv_ophi_data.get(var, np.array([]))
                y_pred = tv_ophi_filtered.get(var, np.zeros_like(y_true))

                # Match lengths
                min_len = min(len(y_true), len(y_pred))
                y_true = y_true[:min_len]
                y_pred = y_pred[:min_len]

                # Avoid zero-division by taking max-abs for scaling
                max_val = max(np.max(np.abs(y_true)), np.max(np.abs(y_pred)))
                if max_val == 0:
                    max_val = 1.0

                # Store the max val
                maxv[sheet_name][var] = max_val

                # Scale data (for the global "scaled" approach to LS)
                y_true_scaled = y_true / max_val
                y_pred_scaled = y_pred / max_val

                # Collect scaled data for overall R² across *all* variables
                global_y_true.setdefault(var, []).extend(y_true_scaled)
                global_y_pred.setdefault(var, []).extend(y_pred_scaled)

                # Collect all points (unscaled) + sigma for weighted metrics
                # Use the standard deviation if available, else 1.0
                sigma_var = std_dev.get(var, 1.0)
                for yt, yp in zip(y_true, y_pred):
                    var_measurements.setdefault(var, []).append((yt, yp, sigma_var))

            # -------------------------------------------------------------
            # Collect measured vs. predicted for TIME-INVARIANT outputs
            # -------------------------------------------------------------
            for var in ti_ophi_vars:
                y_true = ti_ophi_data.get(var, np.nan)
                y_pred = ti_ophi_sim.get(var, np.nan)

                # If either is NaN, skip
                if not np.isnan(y_true) and not np.isnan(y_pred):
                    max_val = max(abs(y_true), abs(y_pred))
                    if max_val == 0:
                        max_val = 1.0

                    maxv[sheet_name][var] = max_val

                    y_true_scaled = y_true / max_val
                    y_pred_scaled = y_pred / max_val

                    global_y_true.setdefault(var, []).append(y_true_scaled)
                    global_y_pred.setdefault(var, []).append(y_pred_scaled)

                    sigma_var = std_dev.get(var, 0)
                    var_measurements.setdefault(var, []).append((y_true, y_pred, sigma_var))

    # -------------------------------------------------------------------------
    # Compute combined metrics
    # -------------------------------------------------------------------------

    # 1) Combine *scaled* data for a global "LS" metric (classical R²):
    all_y_true_scaled = []
    all_y_pred_scaled = []
    for var in global_y_true:
        y_true_list = np.array(global_y_true[var])
        y_pred_list = np.array(global_y_pred[var])
        min_len = min(len(y_true_list), len(y_pred_list))
        y_true_list = y_true_list[:min_len]
        y_pred_list = y_pred_list[:min_len]
        all_y_true_scaled.extend(y_true_list)
        all_y_pred_scaled.extend(y_pred_list)

    all_y_true_scaled = np.array(all_y_true_scaled)
    all_y_pred_scaled = np.array(all_y_pred_scaled)
    min_len = min(len(all_y_true_scaled), len(all_y_pred_scaled))
    all_y_true_scaled = all_y_true_scaled[:min_len]
    all_y_pred_scaled = all_y_pred_scaled[:min_len]

    # Classical R² on the scaled data:
    if len(all_y_true_scaled) > 1:
        ss_res = np.sum((all_y_true_scaled - all_y_pred_scaled) ** 2)
        mean_y_true = np.mean(all_y_true_scaled)
        ss_tot = np.sum((all_y_true_scaled - mean_y_true) ** 2)
        LS = 1 - (ss_res / ss_tot) if ss_tot != 0 else 0.0
    else:
        # Not enough points to compute R²
        LS = np.nan

    # 2) Weighted LS (JWLS) and Weighted MLE across *all* unscaled data:
    #    We'll flatten everything: for each var, we had (y_true, y_pred, sigma).
    #    JWLS = sum( ((y_true - y_pred)/sigma)**2 )
    #    MLE  = sum( 0.5 * [ ln(2π) + ln(sigma^2) + ((y_true - y_pred)/sigma)^2 ] )
    #         = 0.5 * Σ [ ln(2π·sigma_i^2) + ((y_i - yhat_i)/sigma_i)^2 ]
    # Note: This is the negative log-likelihood for i.i.d normal with known sigmas.
    all_residuals_w = []
    sum_mle = 0.0
    sum_chi = 0.0

    # We'll also store R² for each var individually (unweighted, standard definition).
    r2_responses = {}

    for var, measurements in var_measurements.items():
        if len(measurements) == 0:
            continue

        # Separate out arrays
        y_true_arr = np.array([m[0] for m in measurements])
        y_pred_arr = np.array([m[1] for m in measurements])
        sigma_arr   = np.array([m[2] for m in measurements])

        # Weighted residual
        resid = (y_true_arr - y_pred_arr) / sigma_arr
        all_residuals_w.extend(resid)

        # Weighted MLE sum
        # MLE = 0.5 * Σ [ ln(2π sigma^2) + (resid)^2 ]
        # where resid = (y_true - y_pred)/sigma
        # => (resid)^2 = ((y_true - y_pred)/sigma)^2
        # => (y_true - y_pred)^2 / sigma^2
        # We also handle varying sigma_i
        for i in range(len(y_true_arr)):
            sigma_i = sigma_arr[i]
            diff_i = (y_true_arr[i] - y_pred_arr[i])
            sum_mle += 0.5 * (
                np.log(2 * np.pi * sigma_i**2) + (diff_i**2 / sigma_i**2)
            )

        # For Chi-squared
        #   Chi = Σ ( (y_true - y_pred)^2 / |y_pred| ), was unscaled
        #   but in many contexts, "chi-squared" can also be with known sigma
        #   => Σ ( (y_true - y_pred)^2 / sigma^2 ).
        # The user’s code used denominator = abs(y_pred). We'll keep that,
        # but you could adapt if you wanted sigma-based.
        # We'll continue with the user’s definition:
        denom = np.abs(y_pred_arr)
        denom[denom == 0] = 1e-8  # avoid div-zero
        chi_part = np.sum((y_true_arr - y_pred_arr) ** 2 / denom)
        sum_chi += chi_part

        # Compute unweighted R² for *this* var
        if len(y_true_arr) > 1:
            ss_res_var = np.sum((y_true_arr - y_pred_arr)**2)
            mean_var = np.mean(y_true_arr)
            ss_tot_var = np.sum((y_true_arr - mean_var)**2)
            r2_var = 1 - (ss_res_var / ss_tot_var) if ss_tot_var != 0 else 0.0
        else:
            r2_var = np.nan

        r2_responses[var] = r2_var

    # JWLS = Σ residual^2 where residual = (y_true - y_pred) / sigma
    all_residuals_w = np.array(all_residuals_w)
    JWLS = np.sum(all_residuals_w**2)

    # Weighted MLE (already accumulated in sum_mle)
    MLE = sum_mle

    # 3) The original code's "Chi" we keep as well (sum of (res/something)).
    Chi = sum_chi

    # -------------------------------------------------------------------------
    # Store final metrics
    # -------------------------------------------------------------------------
    metrics = {
        'LS' : LS,        # classical R² on scaled data across all variables
        'MLE': MLE,       # negative log-likelihood with known sigmas
        'Chi': Chi,       # as originally defined
        'JWLS': JWLS,     # weighted LS with known sigmas
        'R2_responses': r2_responses  # dict of per-variable R²
    }

    return data, metrics


def _objective_function(theta, data, solver, x0, thetac, objf, thetas, model_structure, run_solver, modelling_settings):
    """
    Wrapper function for the objective function to be used in optimization.

    Parameters:
    theta (list): List of parameters to optimize.
    data (dict): Dictionary containing the experimental data for optimization.
    solver (str): Name of the model to be used.
    x0 (list): Initial guess for the parameters.
    thetac (list): List of parameter scalors.
    nd (int): Number of data points, for time spanning.
    objf (str): Name of the objective function to minimize.
    thetas (list): List indicating which parameters are to be optimized.
    design_variables (dict): Dictionary containing design variables, part of the model structure.
    run_solver (function): Model simulator function.

    Returns:
    float: Value of the objective function.
    """
    # Call the objective function
    optimized_data, metrics = _objective(
        theta, data, [solver], x0, thetac, thetas, model_structure, run_solver, modelling_settings
    )

    # Extract the metrics needed for the optimization
    LS, MLE, Chi, JWLS = metrics['LS'], metrics['MLE'], metrics['Chi'], metrics['JWLS']

    # Determine which objective function to minimize
    if objf == 'LS':
        return -LS
    elif objf == 'MLE':
        return -MLE
    elif objf == 'Chi':
        return -Chi
    elif objf == 'JWLS':
        return JWLS
    else:
        raise ValueError(f"Unknown objective function: {objf}")


def _runner(active_solvers, theta_parameters, bound_max, bound_min, mutation, objf, x0_dict, method, data, model_structure, run_solver, modelling_settings):
    """
    Runner function to perform optimization using different solvers.

    Parameters:
    active_solvers (list): List of active models.
    theta_parameters (dict): Dictionary of theta parameters for each solver.
    bound_max (dict): Dictionary of maximum bounds for parameters of each solver.
    bound_min (dict): Dictionary of minimum bounds for parameters of each solver.
    mutation (dict): Dictionary indicating which parameters are to be optimized for each solver.
    nd2 (int): Number of data points, for time spanning.
    objf (str): Objective function to minimize.
    x0_dict (dict): Dictionary of initial guesses for each solver.
    method (str): Optimization method ('Local' or 'Global').
    data (dict): Dictionary containing the data for optimization.
    design_variables (dict): Dictionary containing design variables.
    run_solver (function): Model simulator function.

    Returns:
    dict: Results of the optimization for each solver.
    """
    results = {}
    x0 = {solver: x0_dict[solver] for solver in active_solvers}

    for solver in active_solvers:
        thetac = theta_parameters[solver]
        thetamax = bound_max[solver]
        thetamin = bound_min[solver]
        thetas = mutation[solver]
        initial_x0 = x0[solver]
        narrow_factor = 1e-40

        bounds = [
            (x * (1 - narrow_factor), x * (1 + narrow_factor)) if not theta else (thetmin, thetmax)
            for x, theta, thetmax, thetmin in zip(initial_x0, thetas, thetamax, thetamin)
        ]

        if method == 'Local':
            result = minimize(
                _objective_function,
                initial_x0,
                args=(data, solver, initial_x0, thetac, objf, thetas, model_structure, run_solver, modelling_settings),
                method='SLSQP',
                bounds=bounds,
                options={'maxiter': 100000, 'disp': False},
                tol=1e-6
            )

        # if method == 'Local':
        #     result = minimize(
        #         _objective_function,
        #         initial_x0,
        #         args=(data, solver, initial_x0, thetac, objf, thetas, model_structure, run_solver, modelling_settings),
        #         method='Nelder-Mead',
        #         options={'maxiter': 100000, 'disp': False, 'fatol': 1e-6}
        #     )


        elif method == 'Global':
            result = differential_evolution(
                _objective_function,
                bounds=bounds,
                args=(data, solver, initial_x0, thetac, objf, thetas, model_structure, run_solver, modelling_settings),
                maxiter=10000,
                popsize=18,
                tol=1e-6,
                strategy='best1bin',
                mutation=(0.5, 1.5),
                recombination=0.7,
                polish=True,
                updating='deferred',
                workers=-1,  # Enable parallel processing
                disp=False
            )

        # Store results for each solver
        results[solver] = {
            'optimization_result': result,
        }
    return results


def _report_optimization_results(results, logging):
    """
    Report the optimization results.

    Parameters:
    results (dict): Dictionary containing the results of the optimization.
    logging (bool): Flag to indicate whether to log the results.
    """
    if logging:
        for solver, solver_results in results.items():
            print(f"Optimization results for {solver}:")
            # print("Optimal Parameters:", solver_results['optimization_result'].x)
            print("Optimization Success:", solver_results['optimization_result'].success)
            print()






def validation(data_storage, model_structure, modelling_settings, estimation_settings, Simula, round_data, framework_settings) -> None: ...
def compute_validation_error(resultun_pred, resultun_val, scaled_params, round_data) -> None: ...



import copy
import pickle
import os
import pandas as pd
from scipy import stats
from middoe.iden_utils import Plotting_Results
from middoe.des_md import MD
from middoe.des_pp import PP
import importlib

def fun_globalizer(func_name):
    """
    Register a proxy function in the kernel_simulator module to call the local function.

    Parameters:
    func_name (str): The name to register the function as.
    """
    kernel_simulator = importlib.import_module('idenpy.kernel_simulator')
    setattr(kernel_simulator, func_name, lambda *args, **kwargs: globals()[func_name](*args, **kwargs))

def run_mbdoe_md(design_settings, model_structure, modelling_settings, core_num, framework_settings, round):
    """
    Run the MBDOE_MD (Model Based Design of Experiments for Model Discrimination).

    Parameters:
    design_settings (dict): User provided - Design settings for the MBDOE.
    model_structure (dict): User provided - The model structure information.
    modelling_settings (dict): User provided - The settings for the modelling process.
    core_num (int): Number of cores to use.
    framework_settings (dict): User provided - Framework settings.
    round (int): The current round of the design - conduction and identification procedure.

    Returns:
    tuple: Design decisions, model discrimination objective function value, and switching points.
    """
    design_settings_copy = copy.deepcopy(design_settings)
    model_structure_copy = copy.deepcopy(model_structure)
    modelling_settings_copy = copy.deepcopy(modelling_settings)
    design_decisions, sum_squared_differences, swps = MD(design_settings_copy, model_structure_copy, modelling_settings_copy, core_num, framework_settings, round)
    return design_decisions, sum_squared_differences, swps

def run_mbdoe_pp(design_settings, model_structure, modelling_settings, core_num, framework_settings, round):
    """
    Run the MBDOE_PP (Model Based Design of Experiments for Parameter Precision).

    Parameters:
    design_settings (dict): User provided - Design settings for the MBDOE.
    model_structure (dict): User provided - The model structure information.
    modelling_settings (dict): User provided - The settings for the modelling process.
    core_num (int): Number of cores to use.
    framework_settings (dict): User provided - Framework settings.
    round (int): The current round of the design - conduction and identification procedure.

    Returns:
    tuple: Design decisions, parameter precision objective function value, and switching points.
    """
    design_settings_copy = copy.deepcopy(design_settings)
    model_structure_copy = copy.deepcopy(model_structure)
    modelling_settings_copy = copy.deepcopy(modelling_settings)
    design_decisions, pp_obj, swps = PP(design_settings_copy, model_structure_copy, modelling_settings_copy, core_num, framework_settings, round)
    return design_decisions, pp_obj, swps


def write_excel(df_combined, experiment_number, excel_path):
    """
    Write the combined DataFrame to an Excel file.

    Parameters:
    df_combined (DataFrame): Combined DataFrame to write.
    experiment_number (int or str): Experiment number to use as the sheet name.
    excel_path (str): Path to the Excel file.
    """
    experiment_number = str(experiment_number)  # Ensure experiment_number is a string
    # Check if the file already exists
    if os.path.isfile(excel_path):
        # Open the file in append mode if it exists
        with pd.ExcelWriter(excel_path, mode='a', engine='openpyxl') as writer:
            existing_sheets = writer.book.sheetnames
            if experiment_number in existing_sheets:
                # Append to the existing sheet
                df_combined.to_excel(writer, sheet_name=experiment_number, index=False)
            else:
                # Create a new sheet
                df_combined.to_excel(writer, sheet_name=experiment_number, index=False)
    else:
        # If the file does not exist, create it and write the data
        with pd.ExcelWriter(excel_path, engine='openpyxl') as writer:
            df_combined.to_excel(writer, sheet_name=experiment_number, index=False)


def read_excel(file_path):
    """
    Read data from an Excel file.

    Parameters:
    file_path (str): Path to the Excel file.

    Returns:
    dict: Data read from the Excel file, with sheet names as keys.
    """
    data = pd.read_excel(file_path, sheet_name=None)
    return data

def save_rounds(j, ranking, k_optimal_value, rCC_values, J_k_values, result, theta_parameters, design_type, round_data, modelling_settings, scaled_params, estimation_settings, solver_parameters, framework_settings, obs, case, data_storage, model_structure):
    """
    Save data for each round of model identification, and append to prior.

    Parameters:
    j (int): Round number, the current round of the design - conduction and identification procedure.
    ranking (list): Ranking of parameters, from estimability analysis.
    k_optimal_value (float): Optimal number of parameters to be estimated, from estimability analysis.
    rCC_values (list): Corrected critical ratios for models.
    J_k_values (list): Objectives of weighted least square method based optimization for models.
    result (dict): Result of the identification procedure (estimation and uncertainty analysis) for models.
    theta_parameters (dict): Parameters for models.
    design_type (str): Type of design (classic or DOE).
    round_data (dict): Dictionary to append the data for the current round.
    modelling_settings (dict): User provided - The settings for the modelling process.
    scaled_params (dict): Estimated parameters scaled to the original scale for each model-round observations.
    estimation_settings (dict): Settings for the estimation process, including active solvers and plotting options.
    solver_parameters (dict): Parameters for the solver.
    framework_settings (dict): User provided - Settings related to the framework, including paths and case information.
    obs (int): Number of observations.
    case (str): Case identifier, used for naming files and directories.
    data_storage (dict): Dictionary to store data of experiments.
    model_structure (dict): User provided - The model structure information.

    Returns:
    dict: Reference t-value for each model-round observation.
    """
    round_key = f'Round {j}'
    dof = {solver: obs - len(theta_parameters[solver]) for solver in modelling_settings['active_solvers']}
    trv = {solver: stats.t.ppf(1 - (1 - 0.95) / 2, dof[solver]) for solver in modelling_settings['active_solvers']}
    round_data[round_key] = {
        'ranking': ranking,
        'k_optimal_value': k_optimal_value,
        'rCC_values': rCC_values,
        'J_k_values': J_k_values,
        'design_type': design_type,
        'mutation': {},
        'original_positions': {},
        'trv': trv,
        'theta_parameters': theta_parameters,
        'scaled_params': scaled_params,
        'result': result,
        'estimation_settings': estimation_settings,
        'framework_settings': framework_settings,
        'model_structure': model_structure
    }

    # Save mutation information
    for solver, mutation in modelling_settings['mutation'].items():
        round_data[round_key]['mutation'][solver] = mutation

    # Save original positions information, if it exists
    for solver in modelling_settings['active_solvers']:
        if solver in modelling_settings.get('original_positions', {}):
            round_data[round_key]['original_positions'][solver] = modelling_settings['original_positions'][solver]
        else:
            round_data[round_key]['original_positions'][solver] = []  # Or handle as needed

    if j == 1:
        add_norm_par(modelling_settings)
    else:
        for solver in modelling_settings['active_solvers']:
            # Initialize 'normalized_parameters' if it doesn't exist
            if 'normalized_parameters' not in modelling_settings:
                modelling_settings['normalized_parameters'] = {}

            # Create the entry for the solver if it doesn't exist
            if solver not in modelling_settings['normalized_parameters']:
                modelling_settings['normalized_parameters'][solver] = []

            # Update the normalized parameters for the solver
            modelling_settings['normalized_parameters'][solver] = result[solver]['estimations_normalized']

    # ranking, k_optimal_value, rCC_values, J_k_values = None, None, None, None
    for solver in modelling_settings['active_solvers']:
        modelling_settings['V_matrix'][solver] = result[solver]['V_matrix']

    solver_cov_matrices = {solver: result[solver]['V_matrix'] for solver in modelling_settings['active_solvers']}
    solver_confidence_intervals = {solver: result[solver]['CI'] for solver in modelling_settings['active_solvers']}
    plotting1 = Plotting_Results(modelling_settings, framework_settings)  # Instantiate Plotting class
    if estimation_settings['con_plot'] == True:
        plotting1.conf_plot(solver_parameters, solver_cov_matrices, solver_confidence_intervals, j)
    if estimation_settings['fit_plot'] == True:
        plotting1.fit_plot(data_storage, result, j, model_structure)

    return trv


def save_to_jac(results, filename):
    """
    Save results (produced from round data saved or screening-sensitivity analysis) to a .jac file.

    Parameters:
    results (dict): Results to save.
    filename (str): Path to the .jac file.
    """
    try:
        with open(filename, 'wb') as file:
            pickle.dump(results, file, protocol=pickle.HIGHEST_PROTOCOL)
        print(f"Results successfully saved to {filename}")
    except Exception as e:
        print(f"An error occurred while saving the results: {e}")

def load_from_jac(filename):
    """
    Load results (produced from round data saved or screening-sensitivity analysis) from a .jac file.

    Parameters:
    filename (str): Path to the .jac file.

    Returns:
    dict or None: Loaded results, or None if the file does not exist.
    """
    # Check if the file exists before attempting to open it
    if os.path.exists(filename):
        with open(filename, 'rb') as file:
            results = pickle.load(file)
        return results
    else:
        print(f"File not found: {filename}")
        return None

def data_appender(df_combined, experiment_number, data_storage):
    """
    Append data for the current experiment to the data storage dictionary
    and return the updated data storage along with the cumulative count of observations.

    Parameters:
    df_combined (DataFrame): Data from the current experiment.
    experiment_number (int or str): The experiment number (used as a key).
    data_storage (dict): Dictionary to store data for each experiment.

    Returns:
    dict: Updated data storage with appended experiment data, and total count of observations.
    """
    # Convert experiment_number to string for consistent dictionary keys
    experiment_number = str(experiment_number)

    # Add the data for this experiment to the storage
    data_storage[experiment_number] = df_combined

    return data_storage


def add_norm_par(modelling_settings):
    """
    Add normalized parameters to the modelling settings.

    Parameters:
    modelling_settings (dict): Dictionary containing modelling settings, including 'theta_parameters'.

    Returns:
    dict: Updated modelling settings with 'normalized_parameters' added.

    Raises:
    KeyError: If 'theta_parameters' is not present in the modelling settings.
    """
    # Check if 'theta_parameters' is present in modelling_settings
    if 'theta_parameters' in modelling_settings:
        # Create 'normalized_parameters' as a dictionary with lists of ones
        modelling_settings['normalized_parameters'] = {
            key: [1] * len(value) for key, value in modelling_settings['theta_parameters'].items()
        }
    else:
        raise KeyError("The dictionary must contain 'theta_parameters' as a key.")

    return modelling_settings



import os
from middoe.iden_utils import validation_R2, validation_params, Plotting_Results
from middoe.iden_parmest import Parmest
from middoe.iden_uncert import Uncert


def validation(data_storage, model_structure, modelling_settings, estimation_settings, Simula, round_data, framework_settings):
    """
    Perform leave-one-out validation to evaluate the generalization of the parameter estimation.

    Parameters:
    data_storage (dict): Experimental data observations.
    model_structure (dict): Structure of the model, including variables and their properties.
    modelling_settings (dict): Settings related to the modelling process, including theta parameters.
    estimation_settings (dict): Settings for the estimation process, including active solvers and plotting options.
    Simula (module): The simulation module used for the experiments.

    Returns:
    dict: A dictionary containing the validation results.
    """
    # Initialize cross-validation results
    sheet_names = list(data_storage.keys())
    # sheet_names = list(data_storage.keys())[1:-1]
    n_sheets = len(sheet_names)
    R2_prd, R2_val, parameters, MSE_pred, MSE_val = {}, {}, {}, {}, {}

    for i in range(n_sheets):
        # Split data into training and validation sets
        validation_sheet = sheet_names[i]
        training_sheets = [sheet for sheet in sheet_names if sheet != validation_sheet]

        training_data = {sheet: data_storage[sheet] for sheet in training_sheets}
        validation_data = {validation_sheet: data_storage[validation_sheet]}

        print(f"Running validation fold {i + 1}/{n_sheets}...")
        print(f"Validation sheet: {validation_sheet}")

        resultpr = Parmest(
            model_structure,
            modelling_settings,
            estimation_settings,
            training_data,
            Simula,
        )

        resultun_pred, theta_parameters_pred, solver_parameters_pred, scaled_params_pred, obs_pred = Uncert(
            training_data,
            resultpr,
            model_structure,
            modelling_settings,
            estimation_settings,
            Simula
        )

        resultun_val, theta_parameters_val, _, _, obs_val = Uncert(
            validation_data,
            resultpr,
            model_structure,
            modelling_settings,
            estimation_settings,
            Simula
        )
        plotting1 = Plotting_Results(modelling_settings, framework_settings)  # Instantiate Plotting class
        plotting1.fit_plot(validation_data, resultun_val, f'val{i+1}', model_structure)
        R2_prd[i+1], R2_val[i+1], parameters[i+1], R2_ref, ref_params, MSE_pred[i+1], MSE_val[i+1], MSE_ref = compute_validation_error(resultun_pred, resultun_val, scaled_params_pred, round_data)

    base_path = framework_settings['path']
    modelling_folder = str(framework_settings['case'])   # No leading backslash here

    # Join the base path and modelling folder
    filename = os.path.join(base_path, modelling_folder)

    # Ensure the 'modelling' directory exists
    os.makedirs(filename, exist_ok=True)
    base_path = filename  # Assuming filename contains the base path
    modelling_folder = 'validation'
    full_path = os.path.join(base_path, modelling_folder)
    os.makedirs(full_path, exist_ok=True)
    validation_R2(R2_prd, R2_val, R2_ref, full_path, case='R2')
    validation_R2(MSE_pred, MSE_val, MSE_ref, full_path, case='MSE')
    validation_params(parameters, ref_params, full_path)


    return R2_prd, R2_val, parameters

# def validation(data_storage, model_structure, modelling_settings, estimation_settings, Simula, round_data, framework_settings):
#     """
#     Perform leave-one-out validation to evaluate the generalization of the parameter estimation.
#
#     Parameters:
#     data_storage (dict): Experimental data observations.
#     model_structure (dict): Structure of the model, including variables and their properties.
#     modelling_settings (dict): Settings related to the modelling process, including theta parameters.
#     estimation_settings (dict): Settings for the estimation process, including active solvers and plotting options.
#     Simula (module): The simulation module used for the experiments.
#
#     Returns:
#     dict: A dictionary containing the validation results.
#     """
#     # Initialize cross-validation results
#     sheet_names = list(data_storage.keys())
#     n_sheets = len(sheet_names)
#     R2_prd, R2_val, parameters, MSE_pred, MSE_val = {}, {}, {}, {}, {}
#
#     # Exclude the first and last sheet from being the validation sheet
#     # for i in range(1, n_sheets - 1):  # Validation sheets range from second to second-to-last
#     for i in range(0, n_sheets):  # Validation sheets range from second to second-to-last
#         validation_sheet = sheet_names[i]
#         training_sheets = [sheet for sheet in sheet_names if sheet != validation_sheet]
#
#         training_data = {sheet: data_storage[sheet] for sheet in training_sheets}
#         validation_data = {validation_sheet: data_storage[validation_sheet]}
#
#         print(f"Running validation fold {i}/{n_sheets - 2}...")
#         print(f"Validation sheet: {validation_sheet}")
#
#         resultpr = Parmest(
#             model_structure,
#             modelling_settings,
#             estimation_settings,
#             training_data,
#             Simula,
#         )
#
#         resultun_pred, theta_parameters_pred, solver_parameters_pred, scaled_params_pred, obs_pred = Uncert(
#             training_data,
#             resultpr,
#             model_structure,
#             modelling_settings,
#             estimation_settings,
#             Simula
#         )
#
#         resultun_val, theta_parameters_val, _, _, obs_val = Uncert(
#             validation_data,
#             resultpr,
#             model_structure,
#             modelling_settings,
#             estimation_settings,
#             Simula
#         )
#
#         plotting1 = Plotting_Results(modelling_settings, framework_settings)  # Instantiate Plotting class
#         plotting1.fit_plot(validation_data, resultun_val, f'val{i}', model_structure)
#
#         R2_prd[i], R2_val[i], parameters[i], R2_ref, ref_params, MSE_pred[i], MSE_val[i], MSE_ref = compute_validation_error(
#             resultun_pred, resultun_val, scaled_params_pred, round_data
#         )
#
#     base_path = framework_settings['path']
#     modelling_folder = str(framework_settings['case'])  # No leading backslash here
#
#     # Join the base path and modelling folder
#     filename = os.path.join(base_path, modelling_folder)
#
#     # Ensure the 'modelling' directory exists
#     os.makedirs(filename, exist_ok=True)
#     base_path = filename  # Assuming filename contains the base path
#     modelling_folder = 'validation'
#     full_path = os.path.join(base_path, modelling_folder)
#     os.makedirs(full_path, exist_ok=True)
#
#     validation_R2(R2_prd, R2_val, R2_ref, full_path, case='R2')
#     validation_R2(MSE_pred, MSE_val, MSE_ref, full_path, case='MSE')
#     validation_params(parameters, ref_params, full_path)
#
#     return R2_prd, R2_val, parameters


def compute_validation_error(resultun_pred, resultun_val, scaled_params, round_data):
    """
    Compute validation error (MSE) between predicted and actual validation results.

    Parameters:
    resultun_pred (dict): Predicted results from the calibration data.
    resultun_val (dict): Validation results from the validation data.

    Returns:
    float: The mean squared error.
    """
    prediction_R2, validation_R2, params, R2_ref, ref_params, prediction_MSE, validation_MSE, MSE_ref = {}, {}, {}, {}, {}, {}, {}, {}
    # Get the last key in round_data
    last_key = list(round_data.keys())[-1]

    for solver in resultun_pred:
        if solver in resultun_val:
            # Extract the 'data' field safely, ensuring it's numeric
            prediction_R2[solver] = resultun_pred[solver]['LS']
            prediction_MSE[solver] = resultun_pred[solver]['MSE']
            validation_R2[solver] = resultun_val[solver]['LS']
            validation_MSE[solver] = resultun_val[solver]['MSE']
            R2_ref[solver] = round_data[last_key]['result'][solver]['LS']
            MSE_ref[solver] = round_data[last_key]['result'][solver]['MSE']
            params = scaled_params
            ref_params[solver] = round_data[last_key]['scaled_params'][solver]

    return prediction_R2, validation_R2, params, R2_ref, ref_params, prediction_MSE, validation_MSE, MSE_ref




import os
from middoe.des_utils import _segmenter, _slicer, _reporter, _par_update, build_var_groups, build_linear_constraints, penalized_objective, constraint_violation
from middoe.krnl_simula import Simula
from collections import defaultdict
from functools import partial
import numpy as np
from scipy.optimize import differential_evolution, minimize, Bounds


def PP(design_settings, model_structure, modelling_settings, core_number, framework_settings, round):
    """
    Perform MBDOE for Parameter Precision (MBDOE_PP) with a structure parallel
    to des-md's MD(...) function.
    """

    # --------------------------- EXTRACT DATA --------------------------- #
    tf = model_structure['t_s'][1]
    ti = model_structure['t_s'][0]

    # Time-variant inputs
    tv_iphi_vars = list(model_structure['tv_iphi'].keys())
    tv_iphi_max = [model_structure['tv_iphi'][var]['max'] for var in tv_iphi_vars]
    tv_iphi_min = [model_structure['tv_iphi'][var]['min'] for var in tv_iphi_vars]
    tv_iphi_seg = [model_structure['tv_iphi'][var]['swp'] for var in tv_iphi_vars]
    tv_iphi_const = [model_structure['tv_iphi'][var]['constraints'] for var in tv_iphi_vars]
    tv_iphi_offsett = [model_structure['tv_iphi'][var]['offsett'] for var in tv_iphi_vars]
    tv_iphi_offsetl = [model_structure['tv_iphi'][var]['offsetl'] for var in tv_iphi_vars]
    tv_iphi_cvp = {var: model_structure['tv_iphi'][var]['design_cvp'] for var in tv_iphi_vars}

    # Time-invariant inputs
    ti_iphi_vars = list(model_structure['ti_iphi'].keys())
    ti_iphi_max = [model_structure['ti_iphi'][var]['max'] for var in ti_iphi_vars]
    ti_iphi_min = [model_structure['ti_iphi'][var]['min'] for var in ti_iphi_vars]

    # Time-variant outputs
    tv_ophi_vars = [
        var for var in model_structure['tv_ophi'].keys()
        if model_structure['tv_ophi'][var].get('measured', True)
    ]
    tv_ophi_seg = [model_structure['tv_ophi'][var]['sp'] for var in tv_ophi_vars]
    tv_ophi_offsett_ophi = [model_structure['tv_ophi'][var]['offsett'] for var in tv_ophi_vars]
    tv_ophi_matching = [model_structure['tv_ophi'][var]['matching'] for var in tv_ophi_vars]

    # Time-invariant outputs
    ti_ophi_vars = [
        var for var in model_structure['ti_ophi'].keys()
        if model_structure['ti_ophi'][var].get('measured', True)
    ]

    # Solver and optimization settings
    active_solvers = modelling_settings['active_solvers']
    estimations = modelling_settings['normalized_parameters']
    ref_thetas = modelling_settings['theta_parameters']
    theta_parameters = _par_update(ref_thetas, estimations)

    # Criterion, iteration, and penalty settings
    design_criteria = design_settings['criteria']['MBDOE_PP_criterion']
    nd = design_settings['iteration_settings']['nd']
    maxpp = design_settings['iteration_settings']['maxpp']
    tolpp = design_settings['iteration_settings']['tolpp']
    eps = design_settings['eps']
    mutation = modelling_settings['mutation']
    V_matrix = modelling_settings['V_matrix']

    # ------------------------ CHOOSE METHOD (Local vs Global) ------------------------ #
    method_key = design_settings['optimization_methods']['mdopt_method']
    if method_key == 'Local':
        result, index_dict = _optimizel(
            tv_iphi_vars, tv_iphi_seg, tv_iphi_max, tv_iphi_min, tv_iphi_const,
            tv_iphi_offsett, tv_iphi_offsetl, tv_iphi_cvp,
            ti_iphi_vars, ti_iphi_max, ti_iphi_min,
            tv_ophi_vars, tv_ophi_seg, tv_ophi_offsett_ophi, tv_ophi_matching,
            ti_ophi_vars,
            tf, ti,
            active_solvers, theta_parameters,
            nd, eps, maxpp, tolpp,
            mutation, V_matrix, design_criteria,
            model_structure, modelling_settings
        )
    elif method_key == 'Global':
        result, index_dict = _optimizeg(
            tv_iphi_vars, tv_iphi_seg, tv_iphi_max, tv_iphi_min, tv_iphi_const,
            tv_iphi_offsett, tv_iphi_offsetl, tv_iphi_cvp,
            ti_iphi_vars, ti_iphi_max, ti_iphi_min,
            tv_ophi_vars, tv_ophi_seg, tv_ophi_offsett_ophi, tv_ophi_matching,
            ti_ophi_vars,
            tf, ti,
            active_solvers, theta_parameters,
            nd, eps, maxpp, tolpp,
            mutation, V_matrix, design_criteria,
            model_structure, modelling_settings
        )
    else:
        raise ValueError(f"Unknown method '{method_key}' for mdopt_method in PP().")

    # ------------------------ RE-EVALUATE AT ND2 ------------------------ #
    nd = design_settings['iteration_settings']['nd2']

    # ------------------- USE FINAL SOLUTION IN _runnerpp ---------------- #
    phi, swps, St, pp_obj, t_values, tv_ophi_dict, ti_ophi_dict, phit = _runner(
        result.x,
        nd,
        tv_iphi_vars, tv_iphi_max,
        ti_iphi_vars, ti_iphi_max,
        tv_ophi_vars, ti_ophi_vars,
        active_solvers, theta_parameters,
        tv_iphi_cvp,
        tf, eps, mutation, V_matrix, design_criteria,
        index_dict,
        model_structure,
        modelling_settings
    )

    # --------------------------- REPORT & SAVE --------------------------- #
    base_path = framework_settings['path']
    modelling_folder = str(framework_settings['case'])
    filename = os.path.join(base_path, modelling_folder)
    os.makedirs(filename, exist_ok=True)

    # You may have a specialized reporter or reuse the same `_reporter`
    phi, phit, swps, St = _reporter(
        phi, phit, swps, St,
        pp_obj,                  # pass the objective for logging
        t_values,
        tv_ophi_dict, ti_ophi_dict,
        tv_iphi_vars, tv_iphi_max,
        ti_iphi_vars, ti_iphi_max,
        tf,
        filename,
        design_criteria,
        round,
        core_number
    )

    # ------------------------- RETURN FINAL DATA ------------------------- #
    design_decisions = {
        'phi': phi,
        'phit': phit,
        'swps': swps,
        'St': St,
        'pp_obj': pp_obj,
        't_values': t_values
    }

    return design_decisions, pp_obj, swps


def _optimizel(tv_iphi_vars, tv_iphi_seg, tv_iphi_max, tv_iphi_min, tv_iphi_const,
                tv_iphi_offsett, tv_iphi_offsetl, tv_iphi_cvp,
                ti_iphi_vars, ti_iphi_max, ti_iphi_min,
                tv_ophi_vars, tv_ophi_seg, tv_ophi_offsett_ophi, tv_ophi_matching,
                ti_ophi_vars,
                tf, ti,
                active_solvers, theta_parameters,
                nd, eps, maxpp, tolpp,
                mutation, V_matrix, design_criteria,
                model_structure, modelling_settings):
    """
    Local optimization for MBDOE_PP using trust-constr (analogous to _optimizel in des-md).
    Returns (result, index_dict).
    """

    # 1) Possibly build a var_groups dict like in des-md
    var_groups = build_var_groups(
        tv_iphi_vars, tv_iphi_offsetl, tv_iphi_offsett, tv_iphi_const,
        tv_ophi_vars, tv_ophi_offsett_ophi
    )

    # 2) Segmenter => bounds, x0, index_pairs, index_dict
    bounds_list, x0, index_pairs_levels, index_pairs_times, index_dict = _segmenter(
        tv_iphi_vars, tv_iphi_seg, tv_iphi_max, tv_iphi_min, tv_iphi_const,
        tv_iphi_offsett, tv_iphi_offsetl,
        ti_iphi_vars, ti_iphi_max, ti_iphi_min,
        tv_ophi_vars, tv_ophi_seg, tv_ophi_offsett_ophi, tv_ophi_matching,
        tf, ti
    )
    lower_bounds = [b[0] for b in bounds_list]
    upper_bounds = [b[1] for b in bounds_list]
    x0 = np.clip(x0, lower_bounds, upper_bounds)
    bounds = Bounds(lower_bounds, upper_bounds)

    # 3) Build linear constraints (similarly to des-md)
    A_level, lb_level, A_time, lb_time, constraints_list = build_linear_constraints(
        x_len=len(x0),
        index_pairs_levels=index_pairs_levels,
        index_pairs_times=index_pairs_times,
        var_groups=var_groups
    )

    # 4) Objective function that calls _runnerpp
    def local_objective(x):
        return pp_objective_function(
            x, nd,
            tv_iphi_vars, tv_iphi_max,
            ti_iphi_vars, ti_iphi_max,
            tv_ophi_vars, ti_ophi_vars,
            active_solvers, theta_parameters,
            tv_iphi_cvp,
            tf, eps, mutation, V_matrix, design_criteria,
            index_dict,
            model_structure,
            modelling_settings
        )

    # 5) trust-constr solve
    result = minimize(
        fun=local_objective,
        x0=x0,
        bounds=bounds,
        constraints=constraints_list,
        method='trust-constr',
        options={
            'maxiter': maxpp,
            'verbose': 3,
            'xtol': tolpp,
            'gtol': tolpp,
            'barrier_tol': 1e-8,
        }
    )

    if not result.success:
        print("Parameter-Precision (Local) Optimization failed!")
    else:
        print("Parameter-Precision (Local) Optimization succeeded!")

    return result, index_dict


def _optimizeg(tv_iphi_vars, tv_iphi_seg, tv_iphi_max, tv_iphi_min, tv_iphi_const,
               tv_iphi_offsett, tv_iphi_offsetl, tv_iphi_cvp,
               ti_iphi_vars, ti_iphi_max, ti_iphi_min,
               tv_ophi_vars, tv_ophi_seg, tv_ophi_offsett_ophi, tv_ophi_matching,
               ti_ophi_vars,
               tf, ti,
               active_solvers, theta_parameters,
               nd, eps, maxpp, tolpp,
               mutation, V_matrix, design_criteria,
               model_structure, modelling_settings):

    # 1) Build var_groups (parallel to des-md)
    var_groups = build_var_groups(
        tv_iphi_vars, tv_iphi_offsetl, tv_iphi_offsett, tv_iphi_const,
        tv_ophi_vars, tv_ophi_offsett_ophi
    )

    # 2) Segmenter => bounds, x0, index_pairs, index_dict
    bounds_list, x0, index_pairs_levels, index_pairs_times, index_dict = _segmenter(
        tv_iphi_vars, tv_iphi_seg, tv_iphi_max, tv_iphi_min, tv_iphi_const,
        tv_iphi_offsett, tv_iphi_offsetl,
        ti_iphi_vars, ti_iphi_max, ti_iphi_min,
        tv_ophi_vars, tv_ophi_seg, tv_ophi_offsett_ophi, tv_ophi_matching,
        tf, ti
    )
    lower_bounds = [b[0] for b in bounds_list]
    upper_bounds = [b[1] for b in bounds_list]
    trust_bounds = Bounds(lower_bounds, upper_bounds)

    # 3) Build linear constraints
    (
        A_level, lb_level,
        A_time, lb_time,
        constraints_list
    ) = build_linear_constraints(
        x_len=len(x0),
        index_pairs_levels=index_pairs_levels,
        index_pairs_times=index_pairs_times,
        var_groups=var_groups
    )

    # 4) Build partial function for local objective (pp_objective_function)
    local_obj_fun = partial(
        pp_objective_function,
        nd=nd,
        tv_iphi_vars=tv_iphi_vars,
        tv_iphi_max=tv_iphi_max,
        ti_iphi_vars=ti_iphi_vars,
        ti_iphi_max=ti_iphi_max,
        tv_ophi_vars=tv_ophi_vars,
        ti_ophi_vars=ti_ophi_vars,
        active_solvers=active_solvers,
        theta_parameters=theta_parameters,
        tv_iphi_cvp=tv_iphi_cvp,
        tf=tf, eps=eps, mutation=mutation,
        V_matrix=V_matrix, design_criteria=design_criteria,
        index_dict=index_dict,
        model_structure=model_structure,
        modelling_settings=modelling_settings
    )

    # 5) Build penalized objective for DE (just like des-md):
    #
    #    First, define arguments for the objective and the constraint violation:
    obj_args = {
        'objective_fun': local_obj_fun
    }
    constraint_args = {
        # If your violation function is named differently, adjust accordingly
        'violation_fun': partial(
            constraint_violation,  # or constraint_violation
            A_level=A_level,
            lb_level=lb_level,
            A_time=A_time,
            lb_time=lb_time
        )
    }

    #    Then define penalized function via partial:
    penalized_de_fun = partial(
        penalized_objective,  # or penalized_objective_de
        obj_args=obj_args,
        constraint_args=constraint_args,
        penalty_weight=1e4  # tune your penalty factor
    )

    # 6) Run Differential Evolution with the penalized objective
    de_bounds = list(zip(lower_bounds, upper_bounds))
    de_result = differential_evolution(
        penalized_de_fun,
        bounds=de_bounds,
        strategy='best1bin',
        maxiter=50,          # or any other suitable number
        popsize=15,
        tol=1e-6,
        mutation=(0.5, 1.0),
        recombination=0.7,
        seed=None,
        polish=False,        # we do our own local refinement
        workers=-1,          # parallel
        updating='deferred',
        disp=True
    )
    x0_de = de_result.x

    # 7) Local refinement with trust-constr
    result = minimize(
        fun=local_obj_fun,
        x0=x0_de,
        bounds=trust_bounds,
        constraints=constraints_list,
        method='trust-constr',
        options={
            'maxiter': maxpp,
            'verbose': 3,
            'xtol': tolpp,
            'gtol': tolpp,
            'barrier_tol': 1e-8,
        }
    )

    if not result.success:
        print("Global+Local Parameter-Precision Optimization failed or did not converge.")
    else:
        print("Global+Local Parameter-Precision Optimization succeeded!")

    return result, index_dict


def pp_objective_function(
    x,
    nd,
    tv_iphi_vars, tv_iphi_max,
    ti_iphi_vars, ti_iphi_max,
    tv_ophi_vars, ti_ophi_vars,
    active_solvers, theta_parameters,
    tv_iphi_cvp,
    tf, eps, mutation, V_matrix, design_criteria,
    index_dict,
    model_structure,
    modelling_settings
):
    """
    Similar to md_objective_function but for PP. We want to maximize the PP-criterion
    => so return -pp_obj from _runnerpp (unless it's a min-type objective).
    """
    try:
        _, _, _, pp_obj, _, _, _, _ = _runner(
            x,
            nd,
            tv_iphi_vars, tv_iphi_max,
            ti_iphi_vars, ti_iphi_max,
            tv_ophi_vars, ti_ophi_vars,
            active_solvers, theta_parameters,
            tv_iphi_cvp,
            tf, eps, mutation, V_matrix, design_criteria,
            index_dict,
            model_structure,
            modelling_settings
        )
        if np.isnan(pp_obj):
            return 1e6
        return -pp_obj  # negative => maximize pp_obj
    except Exception as e:
        print(f"Exception in pp_objective_function: {e}")
        return 1e6


def _runner(
    x,
    nd,
    tv_iphi_vars,
    tv_iphi_max,
    ti_iphi_vars,
    ti_iphi_max,
    tv_ophi_vars,
    ti_ophi_vars,
    active_solvers,
    theta_parameters,
    tv_iphi_cvp,
    tf,
    eps,
    mutation,
    V_matrix,
    MBDOE_PP_criterion,
    index_dict,
    model_structure,
    modelling_settings
):
    """
    Run the simulation and compute the parameter precision criterion for the MBDOE_PP problem.

    Parameters
    ----------
    x : list or array
        Design variables to be optimized.
    nd : int
        Number of time points in the first stage.
    tv_iphi_vars : list
        Time-variant input variables.
    tv_iphi_max : list
        Maximum values for time-variant input variables.
    ti_iphi_vars : list
        Time-invariant input variables.
    ti_iphi_max : list
        Maximum values for time-invariant input variables.
    tv_ophi_vars : list
        Time-variant output variables.
    ti_ophi_vars : list
        Time-invariant output variables.
    active_solvers : list
        List of active solvers.
    theta_parameters : dict
        Dictionary of theta parameters for solvers (keyed by solver_name).
    tv_iphi_cvp : list or dict
        Piecewise interpolation or other design data for time-variant inputs.
    tf : float
        Total experiment time.
    eps : float
        Small epsilon value for numerical sensitivity.
    mutation : dict
        Mutation settings for solvers (keyed by solver_name).
    V_matrix : dict
        V-matrix settings (keyed by solver_name).
    MBDOE_PP_criterion : str
        Criterion for the design (e.g., 'D', 'A', 'E', 'ME').
    index_dict : dict
        Dictionary of index mappings for design variables.
    model_structure : dict
        Model structure dictionary.
    modelling_settings : dict
        Additional solver settings.

    Returns
    -------
    tuple
        (ti, swps, St, pp_obj, t_values, tv_ophi, ti_ophi, phit_interp)
        where:
        - ti (dict): Time-invariant input data.
        - swps (dict): Switching points data.
        - St (dict): Sampling time points (dict of var -> array).
        - pp_obj (float): Performance criterion value.
        - t_values (list): Time values for the simulation.
        - tv_ophi (dict): Time-variant output results (solver_name -> {var -> array}).
        - ti_ophi (dict): Time-invariant output results (solver_name -> {var -> scalar}).
        - phit_interp (dict): Interpolated piecewise data (last solver run).
    """
    # Convert x to list if not already
    x = x.tolist() if not isinstance(x, list) else x

    # ---------------------------------------------------------------------
    # 1) Extract design parameters via _slicer
    # ---------------------------------------------------------------------
    ti, swps, St = _slicer(x, index_dict)
    # Convert St into dict of var -> np.array
    St = {var: np.array(St[var]) for var in St}

    # Combine nd equally spaced points in [0,1] with the sampling points
    t_values_flat = [tp for times in St.values() for tp in times]
    t_values = np.unique(np.concatenate((np.linspace(0, 1, nd), t_values_flat))).tolist()

    # ---------------------------------------------------------------------
    # 2) Prepare data structures
    #    LSA: dict -> dict -> dict, for partial derivatives
    #    J_dot_matrix: dict -> 2D array
    # ---------------------------------------------------------------------
    LSA = defaultdict(lambda: defaultdict(dict))
    J_dot_matrix = defaultdict(
        lambda: np.zeros((len(theta_parameters[active_solvers[0]]),
                          len(theta_parameters[active_solvers[0]])))
    )
    y_values_dict = {}       # y_values_dict[solver_name][var] -> array
    y_i_values_dict = {}     # y_i_values_dict[solver_name][var] -> array
    indices = {}             # indices[var] -> boolean mask for times

    # Check that each tv_ophi_var actually appears in St
    for var in tv_ophi_vars:
        if var not in St:
            raise ValueError(f"No time values found for variable '{var}' in St.")
        # For each var, find the boolean mask that matches St[var] in t_values
        indices[var] = np.isin(t_values, St[var])

    # Matrices to accumulate from partial derivatives
    M0 = defaultdict(
        lambda: np.zeros((len(theta_parameters[active_solvers[0]]),
                          len(theta_parameters[active_solvers[0]])))
    )
    M = {}

    # For returning solver results
    tv_ophi = {}
    ti_ophi = {}
    phit_interp = {}

    # ---------------------------------------------------------------------
    # 3) Main loop over solvers
    # ---------------------------------------------------------------------
    for solver_name in active_solvers:
        thetac = theta_parameters[solver_name]
        # Example: start from a "1.0" vector as your baseline parameter
        theta = np.array([1.0] * len(thetac))

        # Time-invariant and switching points data directly from slicer
        ti_iphi_data = ti
        swps_data = swps

        # Scaling factors for time-invariant (phisc) and time-variant (phitsc) inputs
        phisc = {var: ti_iphi_max[i] for i, var in enumerate(ti_iphi_vars)}
        phitsc = {var: tv_iphi_max[i] for i, var in enumerate(tv_iphi_vars)}
        tsc = tf  # treat tf as a time-scaling factor

        # -----------------------------------------------------------------
        # 3a) Run the solver for unperturbed (baseline) theta
        # -----------------------------------------------------------------
        tv_out, ti_out, interp_data = Simula(
            t_values, swps_data, ti_iphi_data,
            phisc, phitsc, tsc,
            theta, thetac,
            tv_iphi_cvp, {},  # pass empty dict or relevant data
            solver_name,
            model_structure,
            modelling_settings
        )
        tv_ophi[solver_name] = tv_out      # e.g. {var -> array} for time-variant
        ti_ophi[solver_name] = ti_out      # e.g. {var -> scalar} for time-invariant
        phit_interp = interp_data

        # Store unperturbed outputs in y_values_dict
        y_values_dict[solver_name] = {}
        for var in tv_ophi_vars:
            # Make sure it's a NumPy array
            y_values_dict[solver_name][var] = np.array(tv_out[var])
        for var in ti_ophi_vars:
            # Time-invariant output as a scalar or array(1,)
            y_values_dict[solver_name][var] = np.array([ti_out[var]])

        # Create a dictionary for "combined" unperturbed values (by var)
        y_combined = {}
        for var in tv_ophi_vars:
            y_combined[var] = y_values_dict[solver_name][var]
        for var in ti_ophi_vars:
            y_combined[var] = y_values_dict[solver_name][var]

        # -----------------------------------------------------------------
        # 3b) Loop over free parameters (the ones set to True in mutation)
        # -----------------------------------------------------------------
        free_params_indices = [i for i, is_free in enumerate(mutation[solver_name]) if is_free]

        # Initialize dictionary for storing partial-perturbation outputs
        y_i_values_dict[solver_name] = {}

        for para_idx in free_params_indices:
            # Perturb a single parameter
            modified_theta = theta.copy()
            modified_theta[para_idx] += eps

            # -------------------------------------------------------------
            # 3b-i) Run solver with modified theta
            # -------------------------------------------------------------
            tv_out_mod, ti_out_mod, _ = Simula(
                t_values, swps_data, ti_iphi_data,
                phisc, phitsc, tsc,
                modified_theta, thetac,
                tv_iphi_cvp, {},  # pass empty dict or relevant data
                solver_name,
                model_structure,
                modelling_settings
            )

            # Store these "modified" results
            y_i_values_dict[solver_name][var] = {}
            for var in tv_ophi_vars:
                y_i_values_dict[solver_name][var] = np.array(tv_out_mod[var])
            for var in ti_ophi_vars:
                y_i_values_dict[solver_name][var] = np.array([ti_out_mod[var]])

            # Build a dictionary for the perturbed outputs
            y_modified_combined = {}
            for var in tv_ophi_vars:
                y_modified_combined[var] = y_i_values_dict[solver_name][var]
            for var in ti_ophi_vars:
                y_modified_combined[var] = y_i_values_dict[solver_name][var]

            # -------------------------------------------------------------
            # 3b-ii) Compute partial derivatives (LSA)
            #        LSA[var][solver_name][para_idx] = (perturbed - base)/eps
            # -------------------------------------------------------------
            for var in indices:
                LSA[var][solver_name][para_idx] = (
                    y_modified_combined[var] - y_combined[var]
                ) / eps

        # -----------------------------------------------------------------
        # 3c) Construct solver_LSA matrix from LSA
        #     Rows: each var in indices
        #     Cols: each free parameter
        # -----------------------------------------------------------------
        # If "indices" is {var1: mask, var2: mask, ...}, we can iterate over var in sorted(indices).
        # Or keep them in the order they appear. Each row in solver_LSA will be a "stack" of partials
        # for that var. (Alternatively, you might sum them or some other approach.)
        solver_LSA_list = []
        for var in indices:
            row_for_var = [LSA[var][solver_name][i] for i in free_params_indices]
            # row_for_var might be a 1D array or scalar for each i. If you want to combine them into a single float,
            # or if each is an array, you might need to flatten or handle shapes.
            # For now, let's assume each LSA[var][solver_name][i] is a 1D array => we must handle that carefully.
            # A common approach: sum or average those partial derivatives. Or stack them vertically.
            # If you want each var to produce multiple rows, you do something else.
            #
            # For a single row in solver_LSA, we might need to average or something. Let's assume we do a simple sum:
            row_summed = np.array([arr.mean() for arr in row_for_var])  # Example: each arr is a 1D difference
            solver_LSA_list.append(row_summed)
        solver_LSA = np.array(solver_LSA_list)  # shape (#vars, #free_params)

        # Transpose for J_dot = J^T J
        J_transpose = solver_LSA.T  # shape (#free_params, #vars)
        J_dot = np.dot(J_transpose, solver_LSA)  # shape (#free_params, #free_params)

        # Make sure shapes align with J_dot_matrix[solver_name]
        if J_dot_matrix[solver_name].shape != (solver_LSA.shape[1], solver_LSA.shape[1]):
            J_dot_matrix[solver_name] = np.zeros((solver_LSA.shape[1], solver_LSA.shape[1]))

        # Accumulate
        J_dot_matrix[solver_name] += J_dot
        M0[solver_name] = J_dot_matrix[solver_name]

        # Build final M = M0 + inv(V_matrix_for_free_params)
        reduced_V_matrix = V_matrix[solver_name][np.ix_(free_params_indices, free_params_indices)]
        M[solver_name] = M0[solver_name] + np.linalg.inv(reduced_V_matrix)

    # ---------------------------------------------------------------------
    # 4) Evaluate design criterion
    # ---------------------------------------------------------------------
    pp_obj = None
    solver_name = active_solvers[0]  # or whichever solver you want to measure
    if MBDOE_PP_criterion == 'D':
        pp_obj = np.linalg.det(M[solver_name]) * 1e-1
    elif MBDOE_PP_criterion == 'A':
        pp_obj = np.trace(M[solver_name])
    elif MBDOE_PP_criterion == 'E':
        eigenvalues = np.linalg.eigvalsh(M[solver_name])
        pp_obj = np.min(eigenvalues)
    elif MBDOE_PP_criterion == 'ME':
        condition_number = np.linalg.cond(M[solver_name])
        pp_obj = -condition_number

    print(f"\rmbdoe-PP:{MBDOE_PP_criterion} is running with {pp_obj}", end='', flush=True)

    # ---------------------------------------------------------------------
    # 5) Return the relevant pieces
    # ---------------------------------------------------------------------
    # phit_interp might hold data only for the last solver or each solver in a dict
    return ti, swps, St, pp_obj, t_values, tv_ophi, ti_ophi, phit_interp



def MD(design_settings, model_structure, modelling_settings, core_number, framework_settings, round) -> None: ...
def md_objective_function(x, nd, tv_iphi_vars, tv_iphi_max, ti_iphi_vars, ti_iphi_max, tv_ophi_vars, ti_ophi_vars, active_solvers, theta_parameters, tv_iphi_cvp, design_criteria, tf, eps, mutation, index_dict, model_structure, modelling_settings, runner_function) -> None: ...
